#' Survival risk prediction
#'
#' This function implements the survial risk prediction tool in BRB-ArrayTools
#'
#' Please see the BRB-ArrayTools manual for details.
#'
#' @param exprTrain matrix/data frame of gene expression data for training samples. Rows are genes and columns are arrays. Its column names must be provided.
#' @param covTrain matrix/data frame of clinical covariates from the training samples. Rows are arrays and columns are clinical covariates. Default is NULL.
#' @param exprTest matrix/data frame of gene expression data for new samples. Rows are genes and columns are arrays. Its column names must be provided.
#' @param CovTest matrix/data frame of clinical covariates from new samples. Rows are arrays and columns are clinical covariates. Default is NULL.
#' @param geneid matrix/data frame of gene IDs.
#' @param status vector specifying survival status (1 = death, 0 = censored).
#' @param tme vector specifying survival time.
#' @param geneSelect character string of a gene selection method: "pc" for principal component and "pcr" for penalized Cox regression. Default is "pc".
#' @param nriskgroups numeric of the number of risk groups. It can only be 2 or 3. Default is 2.
#' @param progIndexPerc numeric of prognostic index percentile. In 2-risk group case, one percentile values needs to be provided, whereas, in
#' the 3-risk group case, two percentile values need to be provided. Default is 50 for 2-risk group case.
#' @param cvMethod character string of cross validation method. "loocv" for leave-one-out cross validation, and "10fold" for 10-fold cross validation.
#' @param nperm numeric of number of permutation tests. If \code{nperm == 0}, no permutation tests will be conducted; otherwise, permutation tests based
#' on log rank statistics will be run. Default is 0.
#' @param landmarktime numeric of landmark time for the ROC curve. If \code{landmarktime=0}, median of survival time will be used.
#' @param alpha numeric of the threshold significance level of penalized Cox regression model for selecting the genes to
#'              be used in computing the principal components. Default is 0.001.
#' @param ncomp numeric of number of principal components if \code{geneSelect = "pc"}.
#' @param pcrgenes numeric of number of desired genes if \code{geneSelect = "pcr"}.
#' @param mixing numeric of the mixing parameter in penalized Cox regression.
#' @param projectPath character string specifying the full project path.
#' @param outputName character string specifying the output folder name. Default is "SurvivalRiskPrediction".
#' @param generateHTML logical. If \code{TRUE}, an HTML page will be generated with detailed class prediction results
#'                     saved in <projectPath>/Output/<outputName>/<outputName>.html; otherwise not.
#' @return A list that may include the following objects:
#' \itemize{
#'    \item \code{predictNewSamplesModE}: a data frame with risk group predictions of new samples when the model with gene expression only is considered.
#'    \item \code{genesInClassifierModE}: a data frame with genes selected by fitting penalized Cox proportional hazards model
#'                                        when the model with gene expression only is considered.
#'    \item \code{loadingMatrixModE}: a data frame with loading matrix of the significant genes and the correlations between the principal components
#'                                    and the significant genes when the model with gene expression only is considered.
#'    \item \code{predictRiskTrainingModE}: a data frame with Table of arrays, survival time, censoring indicator and survival risk prediction
#'                                          of training samples when the model with gene expression only is considered.
#'    \item \code{predictNewSamplesModCo}: a data frame with risk group predictions of new samples when the model with both covariates and gene expression is considered.
#'    \item \code{genesInClassifierModCo}: a data frame with genes selected by fitting penalized Cox proportional hazards model
#'                                        when the model with both covariates and gene expression is considered.
#'    \item \code{loadingMatrixModCo}: a data frame with loading matrix of the significant genes and the correlations between the principal components
#'                                    and the significant genes when the model with both covariates and gene expression is considered.
#'    \item \code{predictRiskTrainingModCo}: a data frame with Table of arrays, survival time, censoring indicator and survival risk prediction
#'                                          of training samples when the model with both covariates and gene expression is considered.
#'    \item \code{predictNewSamplesModCl}: a data frame with risk group predictions of new samples when the model with covariates only is considered.
#'    \item \code{predictRiskTrainingModCl}: a data frame with Table of arrays, survival time, censoring indicator and survival risk prediction
#'                                          of training samples when the model with covariates only is considered.
#'    \item \code{ppval}: a vector with permutation p-values for the models with gene expression only, covariates only or the combination of gene expression and
#'                        covariates.
#'    \item \code{workPath}: the path for Fortran and other intermediate outputs.
#'    }
#' @export
#' @examples
#' #example 1
#' resList1 <- test.survRiskPredict("Pomeroy")
#' #example2
#' data("Pomeroy")
#' projectPath <- tempdir()
#' outputName <- "SurvivalRiskPrediction"
#' generateHTML <- TRUE
#' resList2 <- survRiskPredict(exprTrain, covTrain, exprTest, covTest,
#'                            geneId, status, tme, geneSelect = "pc",
#'                            nriskgroups = 2, progIndexPerc = 50,
#'                            cvMethod = "10fold", nperm = 0, landmarktime = 0,
#'                            alpha = .001, ncomp = 2, mixing = 1, pcrgenes = 10,
#'                            projectPath = projectPath,
#'                            outputName = outputName, generateHTML)
#' if (generateHTML)
#'   browseURL(file.path(projectPath, "Output", outputName,
#'             paste0(outputName, ".html")))
#' #example3
#' data("Perou")
#' projectPath <- file.path(Sys.getenv("HOME"),"Pomeroy")
#' outputName <- "SurvivalRiskPrediction"
#' generateHTML <- TRUE
#' resList3 <- survRiskPredict(exprTrain, covTrain, exprTest, covTest,
#'                            geneId, status, tme, geneSelect = "pc",
#'                            nriskgroups = 2, progIndexPerc = 50,
#'                            cvMethod = "10fold", nperm = 0, landmarktime = 0,
#'                            alpha = .001, ncomp = 2, mixing = 1, pcrgenes = 10,
#'                            projectPath = projectPath,
#'                            outputName = outputName, generateHTML)
#' if (generateHTML)
#'   browseURL(file.path(projectPath, "Output", outputName,
#'             paste0(outputName, ".html")))

survRiskPredict <- function(exprTrain,
                            covTrain,
                            exprTest = NULL,
                            covTest = NULL,
                            geneid,
                            status,
                            tme,
                            geneSelect = "pc",
                            nriskgroups = 2,
                            progIndexPerc,
                            cvMethod = "10fold",
                            nperm = 0,
                            landmarktime = 0,
                            alpha = 0.001,
                            ncomp = 2,
                            mixing = 1,
                            pcrgenes = 10,
                            projectPath,
                            outputName = "SurvivalRiskPrediction",
                            generateHTML = FALSE) {
  ArrayToolsPath <- path.package("survriskpred")
  WorkPath <- file.path(projectPath, "Fortran", outputName)
  if (!file.exists(WorkPath)) dir.create(WorkPath, recursive = TRUE)

  GenesFiltered <- nrow(exprTrain)
  # TrainingVariable <- ''
  # CovariateVariable <- '' # obtained from VBA
  AnalysisType <- 'SurvivalRisk'
  GeneListFileName1 <- 'SurvivalRiskPrediction_Model1.txt'
  GeneListFileName2 <- 'SurvivalRiskPrediction_Model2.txt'

  unlink(paste(projectPath, '/Genelists/AnalysisResults/', GeneListFileName1, sep = ''))
  unlink(paste(projectPath, '/Genelists/AnalysisResults/', GeneListFileName2, sep = ''))

  # create <experiments.rda> <GeneIdNames.rda>, <GeneIdX.rda> etc.
  KFoldCV <- ifelse(cvMethod == "loocv", 0, 1)
  if (nriskgroups == 2) {
    Percentile1 <- progIndexPerc
  } else {
    Percentile1 <- progIndexPerc[1]
    Percentile2 <- progIndexPerc[2]
  }
  uselasso <- ifelse(geneSelect == "pcr", TRUE, FALSE)
  if (missing(landmarktime) | is.null(landmarktime)) landmarktime <- 0
  prepareSurvRiskPredict(exprTrain, exprTest, geneid, status, tme, covTrain, covTest,
                         nperm, KFoldCV, nriskgroups, Percentile1, Percentile2, WorkPath)
  # note: it does not make sense to write parameters to a file and load them back unless those parameters need to be changed.
  #       so skip 'alpha', 'ncomp', 'uselasso', 'mixing', 'pcrgenes', 'landmarktime' are skipped.

  cat('Getting analysis results ...\n', file=stderr())

  # library(survival)
  DoLOOCV <- FALSE

  load(paste(WorkPath,"/Riskdata.rda",sep=""))  # writeFilesRiskPrediction()

  if (.Machine$sizeof.pointer == 8) {
    dyn.load(paste(ArrayToolsPath,"Fortran/surv64.dll",sep="/"))
  } else {
    dyn.load(paste(ArrayToolsPath,"Fortran/surv.dll",sep="/"))
  }

  if (uselasso && any(tme ==0)) tme[tme == 0] <- .0001 # for a bug in glmnet ver <= 1.5.1. Confirmed from Prof Hastie

  if (missing(landmarktime) | (!missing(landmarktime) && landmarktime <= 0)) landmarktime <- median(tme)
    if (any(is.na(training))) {
    winDialog("ok", "Training variable should not contain NA.")
    stop()
  }

  if (mixing <=0 | mixing >1) {
    winDialog("ok", "Elastic-net mixing parameter must be in the range (0, 1].")
    stop()
  }
  grs.out <- try(resList <- GetRiskResults(projectPath, outputName, ArrayToolsPath, WorkPath, #09/14
                                status, tme, training, arr1, nperm, nfold, clin, progIndexPerc,
                                alpha, ncomp, uselasso, mixing, pcrgenes, landmarktime,
                                GenesFiltered, AnalysisType, generateHTML))
  if (class(grs.out)=='try-error') {
    cat(geterrmessage())
    if (exists('last.warning')) cat(attributes(last.warning)$names)
    stop()
  }

  resList <- c(resList, list(workPath=WorkPath))
  return(resList)
}

prepareSurvRiskPredict <- function(exprTrain, exprTest, geneid, status, tme, covTrain, covTest,
                                   nperm, KFoldCV, nriskgroups, Percentile1, Percentile2, outputpath) {
  # Input:
  #   tme, status: include training and test sets (NA for test sets)
  # Called by survriskpredict()
  # See <WriteFilesRiskPrediction.R>.

  # GeneIdNames.rda (originally created from Utilities.bas)
  GeneIdNames <- colnames(geneid)
  save(GeneIdNames, file = file.path(outputpath, "GeneIdNames.rda"))
  for(i in 1:length(GeneIdNames)) {
    assign(paste0("GeneId", i), geneid[, i])   # NOTE I use a vector instead of one-column data frame
    save(list = paste0("GeneId", i), file = file.path(outputpath, paste0("GeneId", i, ".rda")))
  }

  # experiments.rda and Riskdata.rda (originally from WriteFilesRiskPrediction.R)
  if (is.null(exprTest)) {
    training <- rep(1, length(tme))
  } else {
    training <- c(rep(1, ncol(exprTrain)), rep(0, ncol(exprTest)))
  }
  if (is.null(covTrain)) {
    DoCovariates <- FALSE
  } else if (ncol(covTrain) == 0) {
    DoCovariates <- FALSE
  } else DoCovariates <- TRUE

  # Ting fixed a bug when there are no new samples for risk prediction
  if (!is.null(exprTest))
    writeFilesRiskPrediction(DoCovariates = DoCovariates, covAll = rbind(covTrain, covTest), training, status, tme,
                             KFoldCV, nriskgroups, Percentile1, Percentile2, HasDuplic = FALSE, arr1 = cbind(exprTrain, exprTest),
#							 nperm, Perm = FALSE, duplic = NULL, outputpath) else
                             nperm, duplic = NULL, outputpath) else
     writeFilesRiskPrediction(DoCovariates = DoCovariates, covAll = covTrain, training, status, tme,
                             KFoldCV, nriskgroups, Percentile1, Percentile2, HasDuplic = FALSE, arr1 = exprTrain,
#							 nperm, Perm = FALSE, duplic = NULL, outputpath)
                             nperm, duplic = NULL, outputpath)

}


#' Test survRiskPredict() function
#'
#' This function will load a test dataset and use it run \code{survRiskPredict} function.
#'
#' @param dataset character string specifying one of "Perou" or "Pomeroy" datasets.
#' @param projectPath character string specifying the project path. Default is C:/Users/UserName/Documents/<dataset>.
#' @param outputName character string for the output folder name.
#' @param generateHTML logical. If \code{TRUE}, an html page will be generated with detailed prediction results.
#' @return A list as returned by \code{survRiskPredict}.
#' @export
#' @seealso \code{\link{survRiskPredict}}
#' @examples
#' test.survRiskPredict("Pomeroy")
test.survRiskPredict <- function(dataset = c("Pomeroy", "Perou"), projectPath,
                                 outputName = "SurvivalRiskPrediction", generateHTML = FALSE) {
  dataset <- match.arg(dataset)
  if (missing(projectPath)) {
    projectPath <- file.path(Sys.getenv("R_USER"), dataset)
  }
  # reading data
  expdesign <- read.delim(system.file("extdata", paste0(dataset, "_EXPDESIGN.txt"), package = "survriskpred"), as.is = TRUE)
  geneId <- read.delim(system.file("extdata", paste0(dataset, "_GENEID.txt"), package = "survriskpred"), as.is = TRUE)
  filter <- scan(system.file("extdata", paste0(dataset, "_FILTER.TXT"), package = "survriskpred"), quiet = TRUE)
  singleChannel <- ifelse(dataset == "Pomeroy", TRUE, FALSE)
  if (singleChannel) {
    x <- read.delim(system.file("extdata", paste0(dataset, "_NORMALIZEDLOGINTENSITY.TXT"), package = "survriskpred"), header = FALSE)
    tme <- expdesign[, 7]
    status <- expdesign[, 12]
  } else {
    x <- read.delim(system.file("extdata", paste0(dataset, "_LOGRAT.TXT"), package = "survriskpred"), header = FALSE)
    tme <- expdesign[, 4]
    status <- expdesign[, 5]
  }
  x <- x[filter == 1, ]
  geneId <- geneId[filter == 1, ]

  if(is.null(expdesign$trainsurv)) {
    trainv <- rep("training", length(tme))
  } else {
    trainv <- expdesign$trainsurv
  }

  ind.train <- is.finite(tme) & is.finite(status) & trainv == "training"

  exprTrain <- x[, ind.train]
  colnames(exprTrain) <- expdesign[ind.train, 1]

  ind.test <- trainv == "predict"
  exprTest <- x[, ind.test, drop = FALSE]
  colnames(exprTest) <- expdesign[ind.test, 1]

  if (dataset == "Pomeroy") {
    covTrain <- data.frame(sex = expdesign$Sex[ind.train])
    covTest <- data.frame(sex = expdesign$Sex[ind.test])
  } else {
    covTrain <- NULL
    covTest <- NULL
  }
  #Ting fixed a bug about the order of training and predict samples
  # status and tme should be ordered in the order of training samples
  # followed by new samples 9/28/17
#  status <- status[trainv %in% c("training", "predict")]
#  tme <- tme[trainv %in% c("training", "predict")]
  status <- c(status[trainv=="training"],status[trainv == "predict"])
  tme <- c(tme[trainv=="training"],tme[trainv == "predict"])

  # specify the actual project folder
  resList <- survRiskPredict(exprTrain, covTrain, exprTest, covTest, geneId, status, tme, geneSelect = "pc",
                  nriskgroups = 2, progIndexPerc = 50,
                  cvMethod = "10fold", nperm = 0, landmarktime = 0,
                  alpha = .001, ncomp = 2, mixing = 1, pcrgenes = 10,
                  projectPath = projectPath,
                  outputName = outputName, generateHTML)
  if (generateHTML)
    browseURL(file.path(projectPath, "Output", outputName,
              paste0(outputName, ".html")))

  return(resList)
}
